! This subprogram computes the stationary wealth distribution of the model and computes aggregates
! Written by Enes Sunel
! University of Maryland, March 2009

subroutine Distribution
use parameters
implicit none

!--------------------------------------------------------------------------------------
! TRANSITION RULE OF TOMORROW'S WEALTH AND POLICY FUNCTIONS FOR ASSETS AND CONSUMPTION
!--------------------------------------------------------------------------------------
! Here given our portfolio choice, we will pin down tomorrow's wealth given today's wealth

!!!!--------------------------------------------------------------------------------------
!! Determining the indicator functions for policies
!!!!--------------------------------------------------------------------------------------

do i=1,na
    do j=1,nx

            scalar5 = (1-tau_h)*w*x(j,1)

            I_bond(i,j)   = maxloc(Viter(i,j,:),1)

            b_policy(i,j) = bond(I_bond(i,j),1)

            c_policy(i,j) = Citer(i,j,I_bond(i,j))

            kap_policy(i,j) = c_policy(i,j)**((1-varrho)/scalar3)*scalar6**(varrho/scalar3)

            h_policy(i,j) = ( scalar5/( psi*(1+scalar2*kap_policy(i,j)**gama) ) )**scalar1

            m_policy(i,j) = (c_policy(i,j)/kap_policy(i,j))**(1/varrho)

            a_policy(i,j)   = R_star*b_policy(i,j) + m_policy(i,j)/(1+inf) + Rk_a_x

    end do      ! Closes loop on j =1,nx
end do      ! Closes loop on i =1,na

!-------------------------------------------------------------------------------------------
! LOCATING LAW OF MOTION OF WEALTH IN THE GRID TO FIND THE INVARIANT DISTRIBUTION OF WEALTH
!-------------------------------------------------------------------------------------------

LL = int( (a_policy - alb)/inca ) + 1                           ! Indices of left shoulders

RR = LL + 1                                                     ! Indices of right shoulders


do i=1,na
    do j=1,nx
LW(i,j) = ( asset(RR(i,j),1) - a_policy(i,j) ) / inca           ! Computation of weights for tomorrow's wealth
    end do
end do

RW = 1. - LW

!-----------------------------------------------
! INVARIANT DISTRIBUTION OF WEALTH
!-----------------------------------------------

invasset = 1./(na*nx)

diff=10.

tol1=1.e-06

do
    if (maxval(diff) <= tol1) exit

    trans = 0.

    do i=1,na
        do j=1,nx
                 do jj=1,nx

                trans(LL(i,j),jj) = trans(LL(i,j),jj) + xprob(j,jj)*invasset(i,j)*LW(i,j)

                trans(RR(i,j),jj) = trans(RR(i,j),jj) + xprob(j,jj)*invasset(i,j)*RW(i,j)

                end do
        end do
    end do

    diff = maxval(abs(invasset-trans))

    invasset = trans

end do

print*,'-----------------------------------------------------'
print*,'Sum of pdf over assets: ', sum(invasset)
print*,'-----------------------------------------------------'

!---------------------------------
! AGGREGATE CONSUMPTION
!---------------------------------

C =sum(c_policy*invasset)

!-----------------------------------
! AGGREGATE EFFICIENCY HOURS SUPPLY
!-----------------------------------

H = 0.

do j=1,nx

H = H + sum(h_policy(:,j)*invasset(:,j))*x(j,1)

end do

!---------------------------------
! AGGREGATE BONDS
!---------------------------------

B = sum(b_policy*invasset)

!---------------------------------
! AGGREGATE CAPITAL
!---------------------------------

K = K_a_x

!---------------------------------
! AGGREGATE LABOR DEMAND
!---------------------------------

L = L_a_x

!---------------------------------
! AGGREGATE OUTPUT
!---------------------------------

Y = (K_a_x**alpha*L_a_x**(1.-alpha))**theta

print*,'K_L: ', K/L, 'Y: ', Y
print*,'---------------------------------------------------------------------'

!---------------------------------
! AGGREGATE REAL BALANCES
!---------------------------------

M = sum(m_policy*invasset)

!---------------------------------
! AGGREGATE WEALTH
!---------------------------------

A = sum(a_policy*invasset)

!---------------------------------------
! Aggregate Transactions Costs
!---------------------------------------

TR=0.
do i=1,na
    do j=1,nx

        TR = TR + phi*kap_policy(i,j)**gama*c_policy(i,j)*invasset(i,j)

    end do
end do


!-------------------------------------------------------------------------------------------------------
! Idiosyncratic income in the model is defined as the summation y_disp(i,j) = x(j,1) + (R-1)*bond(i,1) + tau
!The probability of the income level of y(j,i) is invy(j,1)*invbond(i,1)
!-------------------------------------------------------------------------------------------------------

ypdf = 0

do i = 1,na
    do j = 1,nx

        y_disp(i,j)   = (1-tau_h)*w*x(j,1)*h_policy(i,j) + &

                        (1-tau_k)*((K_a_x**alpha*L_a_x**(1.-alpha))**theta - delta*K_a_x - w*L_a_x) + &    ! Disposable income at the idiosyncratic level

                        (R_star -1)*b_policy(i,j) - inf*m_policy(i,j)/(1+inf) + tau

        do t=1,nx
            ypdf(i,j) = ypdf(i,j) + invx(j,1)*invasset(i,t)
        end do

    end do
end do

!---------------------------------
! AGGREGATE INCOME
!---------------------------------
AY = sum(y_disp*ypdf)

print*,'---------------------------------'
!print*,'TOPLAM',sum(ypdf)
print*,'Aggregate Income :',AY
print*,'---------------------------------'


!---------------------------------------
!---------------------------------------
! Checking equilibrium conditions
!---------------------------------------
!---------------------------------------

!-------------------------------
! Implied Public Surplus
!-------------------------------

PS = tau_k*(Y-delta*K-w*L) + tau_h*w*H + inf*M/(1+inf) - G - tau

print*,'AD/Y is: ',(C + TR + delta*K + G + (1-R_star)*B)/Y
print*,'---------------------------------------------------------------------'
print*,'EXCESS DEMAND IN GOODS + LABOR MARKET',(C + TR + delta*K + G + (1.-R_star)*B + w*(L-H) ) - Y
print*,'----------------------------------------------------------------------------'
print*,'PUBLIC SURPLUS', PS
print*,'----------------------------------------------------------------------------'


!!!!------------------------------------------
!!! Updating government spending
!!!!------------------------------------------

if (PS < 0 ) then

g1 = G

G = ( g2 + G )/2.

elseif  (PS > 0 ) then

g2 = G

G = ( g1 + G )/2.

end if

!!!!------------------------------------------
!!! Updating the real wage
!!!!------------------------------------------

H_L = H-L

if ( H_L > 0 ) then     ! excess supply of labor

w1 = w                  ! Update the ES generating real wage

w = ( w2 + w )/2

elseif ( H_L < 0 ) then

w2 = w

w = ( w1 + w )/2

end if


print*,'EXCESS SUPPLY OF LABOR', H_L
print*,'---------------------------------------------------------------------'

!!!!------------------------------------------
!!! Updating the real wage dependent objects
!!!!------------------------------------------

Q = (alpha/(1-alpha))* w / ( (R_star-1)/(1-tau_k) + delta ) ! Homogeneous capital-labor demand ratio implied by the real interest rate, real wage rage and capital income tax rate

L_a_x = ( w/(theta*(1-alpha)*Q**(alpha*theta) )  )**(1/(theta-1)) ! Homogeneous labor demand for the given real wage

K_a_x = Q*L_a_x                                     ! Homogeneous capital demand for the given real wage

print*,'H: ',H,'L_a_x: ',L_a_x,'G: ',G,'w: ',w
print*,'---------------------------------------------------------------------'
print*,'tau/Y: ',tau/Y,'G/Y: ',G/Y,'NX/Y: ',(1-R_star)*B/Y,'Y: ',Y,'C/Y: ',C/Y,'I/Y: ',(delta*K)/Y,'TR/Y: ',&
TR/Y,'SE/Y: ',inf*M/(Y*(1.+inf)),'C/M: ',C/M,'B/(B+M): ',B/(B+M)
print*,'---------------------------------------------------------------------------------------------------'

end subroutine Distribution
