! osl_stat.f90 -- statistics-related procedures
!
! Copyright (C) 2005-2010 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library, a free, open-source
! library for scientific computing, available at http://libosl.org/.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_stat
  use osl_core, wp => osl_wp
  use osl_dist
  implicit none

  interface osl_mean
     module procedure osl_mean_vwp
     module procedure osl_mean_mwp
     module procedure osl_mean_vi
     module procedure osl_mean_mi
  end interface

  interface osl_cov
     module procedure osl_cov_vwp
     module procedure osl_cov_mwp
  end interface

  interface osl_var
     module procedure osl_var_vwp
     module procedure osl_var_mwp
  end interface

contains

  !---------------------------------------------------------------------------!
  ! Sample Statistics                                                         !
  !---------------------------------------------------------------------------!

  !------------------------------------!
  ! Sample mean                        !
  !------------------------------------!

  !! Return the mean of a vector of real numbers V.
  function osl_mean_vwp(v) result(mean)
    real(wp), dimension(:) :: v
    real(wp) :: mean
    mean = sum(v) / real(size(v), wp)
  end function osl_mean_vwp

  !! Return the mean of the rows of a matrix of real numbers A.
  function osl_mean_mwp(a) result(mean)
    real(wp), dimension(:,:) :: a
    real(wp), dimension(size(a,2)) :: mean
    mean = sum(a, DIM=1) / real(size(a,1), wp)
  end function osl_mean_mwp

  !! Return the mean of a vector of integers V.
  function osl_mean_vi(v) result(mean)
    integer, dimension(:) :: v
    real(wp) :: mean
    mean = real(sum(v), wp) / real(size(v), wp)
  end function osl_mean_vi

  !! Return the mean of a matrix of integers A.
  function osl_mean_mi(a) result(mean)
    integer, dimension(:,:) :: a
    real(wp), dimension(size(a,2)) :: mean
    mean = real(sum(a, DIM=1), wp) / real(size(a,1), wp)
  end function osl_mean_mi

  !------------------------------------!
  ! Sample covariance                  !
  !------------------------------------!

  !! Return the sample covariance between two vectors X and Y.  If Y
  !! is omitted, return the variance of X.  The provided means MEAN_X
  !! and MEAN_Y will be used if given, otherwise they will be
  !! calculated internally.
  function osl_cov_vwp(x, y, mean_x, mean_y) result(cov)
    real(wp), dimension(:), intent(in) :: x
    real(wp), dimension(size(x)), intent(in) :: y
    real(wp), intent(in), optional :: mean_x, mean_y
    real(wp) :: cov
    real(wp) :: m_x, m_y
    integer :: i

    if (present(mean_x)) then
       m_x = mean_x
    else
       m_x = osl_mean(x)
    end if

    if (present(mean_y)) then
       m_y = mean_y
    else
       m_y = osl_mean(y)
    end if

    cov = 0.0_wp
    do i = 1, size(x)
       cov = cov + (x(i) - m_x) * (y(i) - m_y)
    end do
    cov = cov / real(size(x) - 1, wp)
  end function osl_cov_vwp

  !! Return the sample covariance between the _rows_ of the matrices X
  !! and Y.  The provided means MEAN_X and MEAN_Y will be used if
  !! given, otherwise they will be calculated internally.
  function osl_cov_mwp(x, y, mean_x, mean_y) result(cov)
    real(wp), dimension(:,:), intent(in) :: x
    real(wp), dimension(size(X,1),size(X,2)), intent(in) :: y
    real(wp), dimension(size(X,2)), intent(in), optional :: mean_x, mean_y
    real(wp), dimension(size(X,2),size(X,2)) :: cov
    real(wp), dimension(size(X,2)) :: m_x, m_y
    integer :: i

    if (present(mean_x)) then
       m_x = mean_x
    else
       m_x = osl_mean(x)
    end if

    if (present(mean_y)) then
       m_y = mean_y
    else
       m_y = osl_mean(y)
    end if

    cov = 0.0_wp
    do i = 1, size(X,1)
       cov = cov + osl_outer_product(X(i,:) - m_x, Y(i,:) - m_y)
    end do
    cov = cov / real(size(X,1) - 1, wp)
  end function osl_cov_mwp

  !------------------------------------!
  ! Sample variance                    !
  !------------------------------------!

  !! Return the sample variance of a vector X.  The provided MEAN will
  !! be used if given, otherwise it will be calculated internally.
  function osl_var_vwp(x, mean) result(var)
    real(wp), dimension(:), intent(in) :: x
    real(wp), intent(in), optional :: mean
    real(wp) :: var, m

    if (.not. present(mean)) then
       m = osl_mean(x)
    else
       m = mean
    end if

    var = osl_cov(x, x, m, m)
  end function osl_var_vwp

  !------------------------------------!
  ! Combined sample statistics         !
  !------------------------------------!

  !! Return the sample covariance between the rows of the matrices X
  !! and Y.  If Y is omitted, return the variance of X.  The provided
  !! means will be used if given, otherwise they will be calculated
  !! internally.
  function osl_var_mwp(x, mean) result(var)
    real(wp), dimension(:,:), intent(in) :: x
    real(wp), dimension(size(X,2)), intent(in), optional :: mean
    real(wp), dimension(size(X,2),size(X,2)) :: var
    real(wp), dimension(size(X,2)) :: m

    if (.not. present(mean)) then
       m = osl_mean(x)
    else
       m = mean
    end if

    var = osl_cov(x, x, m, m)
  end function osl_var_mwp

  !! Calculate sample statistics for a sample of vectors.  The
  !! observations must be stored as rows of the variable SAMPLE.  Any
  !! of the following statistics may be calculated (all are optional):
  !! mean (MEAN), standard deviation (SD), variance (VAR), bias
  !! (BIAS), mean squared error (MSE), root mean squared error (RMSE).
  !! Note that TRUTH must be present in order to calculate BIAS, MSE,
  !! or RMSE (zero is returned otherwise).
  subroutine osl_sample_stats(sample, mean, var, sd, truth, bias, mse, rmse)
    real(wp), dimension(:,:), intent(in) :: sample
    real(wp), dimension(size(sample,2)), optional, intent(out) :: mean
    real(wp), dimension(size(sample,2)), optional, intent(out) :: var
    real(wp), dimension(size(sample,2)), optional, intent(out) :: sd
    real(wp), dimension(size(sample,2)), optional, intent(in) :: truth
    real(wp), dimension(size(sample,2)), optional, intent(out) :: bias
    real(wp), dimension(size(sample,2)), optional, intent(out) :: mse
    real(wp), dimension(size(sample,2)), optional, intent(out) :: rmse

    real(wp), dimension(size(sample,2)) :: m
    real(wp), dimension(size(sample,2)) :: v
    integer :: i

    m = osl_mean(sample)
    if (present(mean)) then
       mean = m
    end if

    ! Calculate the component-wise variance, not the covariance matrix
    v = 0.0_wp
    do i = 1, size(sample, 1)
       v = v + (sample(i,:) - m)**2
    end do
    v = v / real(size(sample, 1) - 1, wp)
    if (present(var)) then
       var = v
    end if

    if (present(sd)) then
       sd = sqrt(v)
    end if

    if (present(truth)) then
       if (present(bias)) then
          bias = m - truth
       end if
       if (present(mse)) then
          if (present(bias)) then
             mse = v + bias**2
          else
             mse = v + (m - truth)**2
          end if
       end if
       if (present(rmse)) then
          if (present(mse)) then
             rmse = sqrt(mse)
          else
             if (present(bias)) then
                rmse = sqrt(v + bias**2)
             else
                rmse = sqrt(v + (m - truth)**2)
             end if
          end if
       end if
    else
       if (present(bias)) bias = 0.0_wp
       if (present(mse)) mse = 0.0_wp
       if (present(rmse)) rmse = 0.0_wp
    end if
  end subroutine osl_sample_stats

  !---------------------------------------------------------------------------!
  ! Tauchen's Procedure                                                       !
  !---------------------------------------------------------------------------!

  !! Construct a discrete Markov chain approximation for a given AR(1)
  !! process using Tauchen's (1986) method.  This code is based on code from
  !! the online appendix of Aruoba, Fernandez-Villaverde, and Rubio-Ramirez
  !! (2006).
  !!
  !! *References*
  !!
  !!  - Tauchen, G. (1986). Finite State Markov-Chain Approximations to
  !!    Univariate and Vector Autoregressions.  _Economics Letters_ 20,
  !!    177-181.
  !!
  !!  - Aruoba, S. B., J. Fernandez-Villaverde, and J. F. Rubio-Ramirez
  !!    (2006). Comparing Solution Methods for Dynamic Equilibrium
  !!    Economies. _J. Economic Dynamics and Control_ 30, 2477-2508.
  subroutine osl_tauchen(rho, sigma, n, cover, z, pi)
    real(wp), intent(in) :: rho                 ! autoregressive coefficient
    real(wp), intent(in) :: sigma               ! variance of Normal term
    integer, intent(in) :: n                    ! number of support points
    real(wp), intent(in) :: cover               ! coverage percentage
    real(wp), dimension(n), intent(out) :: z    ! Markov Chain states
    real(wp), dimension(n,n), intent(out) :: pi ! transition matrix

    ! Local variables
    integer  :: i, j
    real(wp) :: sigma_z, mu, upp, low

    ! Standard deviation of z
    sigma_z = sigma / sqrt(1.0-rho**2)

    ! Define the discrete states of the Markov chain
    z(n) = cover * sigma_z
    z(1) = -z(n)
    do i = 2, n - 1
       z(i) = z(1) + (z(n) - z(1)) * real(i - 1, wp) / real(n - 1, wp)
    end do

    ! Compute the transition matrix
    do i = 1, n
       mu = rho * z(i)
       upp = (z(2) + z(1)) / 2.0_wp
       upp = (upp - mu) / sigma
       pi(i,1) = osl_normal_cdf(upp)
       low = (z(n) + z(n-1)) / 2.0_wp
       low = (low - mu) / sigma
       pi(i,n) = 1.0_wp - osl_normal_cdf(low)
       do j = 2, n - 1
          low = (z(j-1) + z(j)) / 2.0_wp
          upp = (z(j+1) + z(j)) / 2.0_wp
          low = (low - mu) / sigma
          upp = (upp - mu) / sigma
          pi(i,j) = osl_normal_cdf(upp) - osl_normal_cdf(low)
       end do
    end do
  end subroutine osl_tauchen

end module osl_stat
