! osl_core.f90 -- OSL precision, constants, and core procedures
!
! Copyright (C) 2005-2010 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library, a free, open-source
! library for scientific computing, available at http://libosl.org/.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_core
  implicit none

  public

  ! version information
  character(len=*), parameter :: OSL_VERSION = '4.0'
  integer, parameter :: OSL_MAJOR_VERSION = 0
  integer, parameter :: OSL_MINOR_VERSION = 0
  integer, parameter :: OSL_MICRO_VERSION = 0

  ! real precision constants
  integer, parameter :: osl_sp = selected_real_kind(6, 37)
  integer, parameter :: osl_dp = selected_real_kind(15, 307)
! integer, parameter :: osl_qp = selected_real_kind(33, 4931)
  integer, parameter :: osl_wp = osl_dp

  ! integer precision constants
  integer, parameter :: osl_i1b = selected_int_kind(2)
  integer, parameter :: osl_i2b = selected_int_kind(4)
  integer, parameter :: osl_i4b = selected_int_kind(9)
  integer, parameter :: osl_i8b = selected_int_kind(18)

  ! internal-use private precision constants
  integer, parameter, private :: sp = osl_sp
  integer, parameter, private :: dp = osl_dp
  integer, parameter, private :: wp = osl_wp

  ! precision constants
  real(wp), parameter :: osl_eps_sp = epsilon(1.0_sp)
  real(wp), parameter :: osl_eps_dp = epsilon(1.0_dp)
  real(wp), parameter :: osl_eps_wp = epsilon(1.0_wp)

  ! mathematical constants
  real(wp), parameter :: &
       OSL_E        = 2.7182818284590452353602874713526625_wp, & ! e
       OSL_LOG2E    = 1.4426950408889634073599246810018921_wp, & ! log_2 e
       OSL_LOG10E   = 0.4342944819032518276511289189166051_wp, & ! log_10 e
       OSL_LN2      = 0.6931471805599453094172321214581766_wp, & ! ln 2
       OSL_LN10     = 2.3025850929940456840179914546843642_wp, & ! ln 10
       OSL_LNPI     = 1.1447298858494001741434273513530587_wp, & ! ln pi
       OSL_PI       = 3.1415926535897932384626433832795029_wp, & ! pi
       OSL_PI_2     = 1.5707963267948966192313216916397514_wp, & ! pi / 2
       OSL_PI_4     = 0.7853981633974483096156608458198757_wp, & ! pi / 4
       OSL_2PI      = 6.2831853071795864769252867665590058_wp, & ! 2 * pi
       OSL_1_PI     = 0.3183098861837906715377675267450287_wp, & ! 1/pi
       OSL_2_PI     = 0.6366197723675813430755350534900574_wp, & ! 2/pi
       OSL_2_SQRTPI = 1.1283791670955125738961589031215452_wp, & ! 2/sqrt(pi)
       OSL_SQRT1_2  = 0.7071067811865475244008443621048490_wp, & ! 1/sqrt(2)
       OSL_SQRT2    = 1.4142135623730950488016887242096981_wp, & ! sqrt(2)
       OSL_SQRTPI   = 1.7724538509055160272981674833411452_wp, & ! sqrt(pi)
       OSL_SQRT3    = 1.7320508075688772935274463415058724_wp, & ! sqrt(3)
       OSL_SQRT5    = 2.2360679774997896964091736687312762_wp, & ! sqrt(5)
       OSL_EULER    = 0.5772156649015328606065120900824024_wp, & ! gamma
       OSL_PHI      = 1.6180339887498948482045868343656381_wp    ! (1+sqrt(5))/2

  ! default edit descriptors
  character(len=*), parameter :: osl_ed_i   = 'i10'
  character(len=*), parameter :: osl_ed_wp  = 'g17.5'

  !! A context mold for generic interfaces.
  integer, dimension(1), parameter :: osl_ctx = (/ 0 /)

  ! error handling behavior
  logical, private :: osl_error_halt = .true.
  integer, parameter :: osl_error_msg_len = 64

  ! status/error codes
  integer, parameter :: OSL_FAIL     = -1
  integer, parameter :: OSL_SUCCESS  = 0
  integer, parameter :: OSL_EDOM     = 1   ! input domain error
  integer, parameter :: OSL_ERANGE   = 2   ! output range error
  integer, parameter :: OSL_EINVAL   = 3   ! invalid argument
  integer, parameter :: OSL_ESANITY  = 7   ! sanity check failed
  integer, parameter :: OSL_ENOMEM   = 8   ! memory allocation failed
  integer, parameter :: OSL_EMAXITER = 11  ! exceeded maximum iterations
  integer, parameter :: OSL_EUNDRFLW = 15  ! underflow
  integer, parameter :: OSL_EOVRFLW  = 16  ! overflow
  integer, parameter :: OSL_EBADLEN  = 19  ! dimensions are non-conforming
  integer, parameter :: OSL_ENOTSQR  = 20  ! matrix is not square
  integer, parameter :: OSL_EINIT    = 100 ! object not initialized

  ! Procedures for printing informative messages
  interface osl_print
     module procedure osl_print_str
     module procedure osl_print_swp
     module procedure osl_print_vwp
     module procedure osl_print_mwp
     module procedure osl_print_si
     module procedure osl_print_vi
     module procedure osl_print_mi
  end interface

  ! Miscellaneous Mathematical Functions
  interface osl_sgn
     module procedure osl_sgn_i
     module procedure osl_sgn_wp
  end interface

  ! Logged sum of exponentials
  interface osl_log_sum_exp
     module procedure osl_log_sum_exp_vwp
     module procedure osl_log_sum_exp_mwp
  end interface

  ! IEEE floating point procedures
  interface osl_is_finite
     module procedure osl_is_finite_wp
  end interface

  interface osl_is_nan
     module procedure osl_is_nan_wp
  end interface

  ! Swap
  interface osl_swap
     module procedure osl_swap_scalar_wp
     module procedure osl_swap_scalar_i
     module procedure osl_swap_vector_wp
     module procedure osl_swap_vector_i
  end interface

  ! Reverse
  interface osl_reverse
     module procedure osl_reverse_vector_wp
     module procedure osl_reverse_vector_i
  end interface

  ! Linear spaced vectors
  interface osl_linspace
     module procedure osl_linspace_swp
     module procedure osl_linspace_vwp
  end interface

contains


  !---------------------------------------------------------------------------!
  ! Version information                                                       !
  !---------------------------------------------------------------------------!

  !! Checks the version of the OSL.  Returns `.true.` if the
  !! version of the library is at least as new as the given
  !! version and `.false.` otherwise.
  function osl_check_version(major, minor, micro)
    integer, intent(in) :: major, minor, micro
    logical :: osl_check_version
    osl_check_version = &
         (OSL_MAJOR_VERSION >= major) .or. &
         (OSL_MAJOR_VERSION == major .and. OSL_MINOR_VERSION >= minor) .or. &
         (OSL_MAJOR_VERSION == major .and. OSL_MINOR_VERSION == minor .and. &
          OSL_MICRO_VERSION >= micro)
  end function osl_check_version

  !--------------------------------------------------------------------------!
  ! Formatted output                                                         !
  !--------------------------------------------------------------------------!

  !------------------------------------!
  ! Informative messages               !
  !------------------------------------!

  !! Print a message alongside a `real(wp)` matrix.
  subroutine osl_print_mwp(msg, A)
    character(len=*), intent(in) :: msg
    real(wp), dimension(:,:), intent(in) :: A
    integer :: i

    print '(a)', msg
    do i = 1, size(A,1)
       print '(100' // osl_ed_wp // ')', A(i,:)
    end do
  end subroutine osl_print_mwp

  !! Print a message alongside an `integer` matrix.
  subroutine osl_print_mi(msg, A)
    character(len=*), intent(in) :: msg
    integer, dimension(:,:), intent(in) :: A
    integer :: i

    print '(a)', msg
    do i = 1, size(A,1)
       print '(100' // osl_ed_i // ')', A(i,:)
    end do
  end subroutine osl_print_mi

  !! Print a message above a `real(wp)` vector.  If the vector has more than
  !! ten elements, then ten elements will be printed on each line.
  subroutine osl_print_vwp(msg, vec)
    character(len=*), intent(in) :: msg
    real(wp), dimension(:), intent(in) :: vec
    character(len=25) :: fmt_str
    integer :: i, ll, ncols, lines

    print '(a)', msg

    fmt_str = '(10' // osl_ed_wp // ')'

    ncols = size(vec, 1)
    ll = 0
    if (ncols > 10) then
       lines = int(ncols/10.0_wp)

       do i = 1, lines
          ll = 10 * (i - 1)
          print fmt_str, vec(1+ll:10+ll)
       end do
       print fmt_str, vec(11+ll:ncols)
    else
       print fmt_str, vec
    end if
  end subroutine osl_print_vwp

  !! Print a message above an `integer` vector.  If the vector has more
  !! than ten elements, ten elements will be printed on each line.
  subroutine osl_print_vi(msg, vec)
    character(len=*), intent(in) :: msg
    integer, dimension(:), intent(in) :: vec
    character(len=25) :: fmt_str
    integer :: i, ll, ncols, lines

    print '(a)', msg

    fmt_str = '(10' // osl_ed_i // ')'
    ncols = size(vec, 1)
    ll = 0
    if (ncols > 10) then
       lines = int(ncols/10.0_wp)

       do i = 1, lines
          ll = 10 * (i - 1)
          print fmt_str, vec(1+ll:10+ll)
       end do
       print fmt_str, vec(11+ll:ncols)
    else
       print fmt_str, vec
    end if
  end subroutine osl_print_vi

  !! Print a message alongside a `real(wp)` scalar.
  subroutine osl_print_swp(msg, x)
    character(len=*), intent(in) :: msg
    real(wp), intent(in) :: x
    print '(a,' // osl_ed_wp // ')', msg, x
  end subroutine osl_print_swp

  !! Print a message alongside an `integer` scalar.
  subroutine osl_print_si(msg, x)
    character(len=*), intent(in) :: msg
    integer, intent(in) :: x
    print '(a,' // osl_ed_i // ')', msg, x
  end subroutine osl_print_si

  !! Prints a single string.
  subroutine osl_print_str(msg)
    character(len=*), intent(in) :: msg
    print '(a)', msg
  end subroutine osl_print_str

  !------------------------------------!
  ! Headings                           !
  !------------------------------------!

  ! Print a heading consisting of TEXT underlined by equals signs.
  subroutine osl_print_heading(text)
    character(len=*), intent(in) :: text
    print '(/,a,/,a,/,/)', text, repeat("=", len_trim(text))
  end subroutine osl_print_heading

  ! Print a subheading consisting of TEXT underlined by hyphens.
  subroutine osl_print_subheading(text)
    character(len=*), intent(in) :: text
    print '(/,a,/,a,/,/)', text, repeat("-", len_trim(text))
  end subroutine osl_print_subheading

  !--------------------------------------------------------------------------!
  ! Error Handling                                                           !
  !--------------------------------------------------------------------------!

  !! Configure OSL error handling behavior.
  subroutine osl_error_behavior(halt)
    logical, intent(in), optional :: halt

    if (present(halt)) then
       osl_error_halt = halt
    end if
  end subroutine osl_error_behavior

  !! Default error handler.  If `osl_error_halt` is `.true.`, print
  !! the error message and terminate the program.  Otherwise, the
  !! calling program is expected to handle the error.
  subroutine osl_error(reason, errno)
    character(len=*), intent(in) :: reason
    integer, intent(in) :: errno

    if (osl_error_halt) then
       print "('ERROR: ', a, ' (', i2.2, ')')", reason, errno
       stop
    end if
  end subroutine osl_error

  !! Return a string describing the error with number `errno`.
  function osl_error_msg(errno)
    integer, intent(in) :: errno
    character(len=osl_error_msg_len) :: osl_error_msg

    select case (errno)
    case (OSL_FAIL)
       osl_error_msg = 'failure'
    case (OSL_SUCCESS)
       osl_error_msg = 'success'
    case (OSL_EDOM)
       osl_error_msg = 'input domain error'
    case (OSL_ERANGE)
       osl_error_msg = 'output range error'
    case (OSL_EINVAL)
       osl_error_msg = 'invalid argument'
    case (OSL_ESANITY)
       osl_error_msg = 'sanity check failed'
    case (OSL_ENOMEM)
       osl_error_msg = 'memory allocation failed'
    case (OSL_EMAXITER)
       osl_error_msg = 'exceeded maximum number of iterations'
    case (OSL_EUNDRFLW)
       osl_error_msg = 'underflow'
    case (OSL_EOVRFLW)
       osl_error_msg = 'overflow'
    case (OSL_EBADLEN)
       osl_error_msg = 'dimensions are non-conforming'
    case (OSL_ENOTSQR)
       osl_error_msg = 'matrix is not square'
    case default
       osl_error_msg = 'unknown error'
    end select
  end function osl_error_msg

  !--------------------------------------------------------------------------!
  ! Miscellaneous Mathematical Functions                                     !
  !--------------------------------------------------------------------------!

  !------------------------------------!
  ! The Signum Function                !
  !------------------------------------!

  !! Return 1 if the X is nonnegative and -1 otherwise.
  elemental function osl_sgn_i(x)
    integer, intent(in) :: x
    integer :: osl_sgn_i
    if (x < 0) then
       osl_sgn_i = -1
    else
       osl_sgn_i = 1
    end if
  end function osl_sgn_i

  !! Return 1 if X is nonnegative and -1 otherwise.
  elemental function osl_sgn_wp(x)
    real(wp), intent(in) :: x
    integer :: osl_sgn_wp
    if (x < 0.0_wp) then
       osl_sgn_wp = -1
    else
       osl_sgn_wp = 1
    end if
  end function osl_sgn_wp

  !------------------------------------!
  ! Logged sum of exponentials         !
  !------------------------------------!

  !! Given a matrix `V = ( v(k,j) )` for `k = 1, ..., K` and `j = 1,...,J`, this
  !! function returns a vector of length `K` where the `k`-th component is
  !! `log( exp( v(k,1) ) + ... + exp( v(k,J) ) )`.  This function calls
  !! the real(wp) vector version of this function for each row of the matrix `V`.
  !!
  !! Here, V is a K-by-J input matrix and E is a K-by-1 output vector.
  function osl_log_sum_exp_mwp(V) result(E)
    real(wp), dimension(:,:), intent(in) :: V
    real(wp), dimension(size(V,1)) :: E

    integer :: k

    do k = 1, size(V,1)
       E(k) = osl_log_sum_exp_vwp(V(k,:))
    end do
  end function osl_log_sum_exp_mwp

  !! Given a vector `v` of length `J`, this function returns the value
  !! `log(exp(v(1)) + ... + exp(v(J)))`.  In order to prevent numerical
  !! overflow, it first shifts the vector `v` by a constant `c`,
  !! applies `exp` and `log`, and then adjusts the result to account
  !! for the shift.
  !!
  !! Consider the case when `v = (a, b)`. Note that we can write
  !! `exp(a) + exp(b) = [exp(a-c) + exp(b-c)]*exp(c)` for any `c`.
  !! Furthermore, we have
  !! `log( ( exp(a-c) + exp(b-c) ) * exp(c) ) = log( exp(a-c) + exp(b-c) ) + c`.
  !! By choosing `c` appropriately, we can attempt to ensure that `exp` is not
  !! evaluated at a number large enough to cause overflow.  Underflow is also
  !! possible as taking `log` of a number close to zero may result in `-Inf`.
  !! Thus, this function also adjusts for large negative elements of `v`.  Note that
  !! this function is still not completely robust to very poorly scaled vectors `v`.
  !! Overflow or underflow is still possible.
  !!
  !! For a set of `J` constants `u_1, ..., u_J`, if `e_j` is _iid_ with Type 1
  !! Extreme value distribution, this function returns the expected value of
  !! `max(u_1 + e_1, ..., u_J + e_J)`.
  function osl_log_sum_exp_vwp(v) result(e)
    real(wp), dimension(:), intent(in) :: v
    real(wp) :: c, e

    ! Choose c to be the element of v that is largest in absolute value.
    if ( maxval(abs(v)) > maxval(v) ) then
       c = minval(v)
    else
       c = maxval(v)
    end if
    e = log(sum(exp(v-c))) + c
  end function osl_log_sum_exp_vwp

  !------------------------------------!
  ! Sigmoid Function                   !
  !------------------------------------!

  !! Takes as input a vector of length `n`-1 and passes it through a
  !! multidimensional sigmoid mapping to return a vector of length _n_ whose
  !! elements are nonnegative and sum to one.  The `i`-th element of the
  !! input vector controls the relative magnitude of the `i`-th element of
  !! the output vector.  The value corresponding to the `n`-th element is
  !! internally normalized to 1.  Thus, if input element `j` is smaller than 1,
  !! then on output element `j` will be smaller than element _n_.
  !! Similarly if input element `j` is larger than 1.
  !!
  !! This function is useful for performing unconstrained optimization on
  !! parameters which correspond to a complete set of probabilities, since
  !! each of them must be nonnegative and the set must sum to one.
  function osl_sigmoid(theta)
    real(wp), dimension(:), intent(in) :: theta
    real(wp), dimension(size(theta,1)+1) :: osl_sigmoid
    real(wp), dimension(size(theta,1)) :: exptheta
    integer :: n

    n = size(theta,1) + 1
    exptheta = exp(theta)

    osl_sigmoid(1:n-1) = exptheta
    osl_sigmoid(n) = 1.0_wp
    osl_sigmoid = osl_sigmoid/(1.0_wp + sum(exptheta)) ! Normalize
  end function osl_sigmoid

  !--------------------------------------------------------------------------!
  ! IEEE Floating Point Arithmetic                                           !
  !--------------------------------------------------------------------------!

  !! Return `.true.` if the value of the given real variable is `NaN` and
  !! `.false.` otherwise.
  elemental function osl_is_nan_wp(x)
    real(wp), intent(in):: x
    logical :: osl_is_nan_wp

    osl_is_nan_wp = .true.
    if (x >= 0.0_wp) then
       osl_is_nan_wp = .false.
    else if (x <= 0.0_wp) then
       osl_is_nan_wp = .false.
    end if
  end function osl_is_nan_wp

  !! Return `.true.` if the given real variable is finite and `.false.`
  !! otherwise.  Finite is defined as value other than `Inf` or `NaN`.
  elemental function osl_is_finite_wp(x)
    real(wp), intent(in):: x
    logical :: osl_is_finite_wp

    osl_is_finite_wp = .not. (osl_is_nan(x) .or. abs(x) > huge(x))
  end function osl_is_finite_wp

  !--------------------------------------------------------------------------!
  ! Swap                                                                     !
  !--------------------------------------------------------------------------!

  !! Swap the values of two `real(wp)` scalars.
  subroutine osl_swap_scalar_wp(a, b)
    real(wp), intent(inout) :: a, b
    real(wp) :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_scalar_wp

  !! Swap the values of two `integer` scalars.
  subroutine osl_swap_scalar_i(a, b)
    integer, intent(inout) :: a, b
    integer :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_scalar_i

  !! Swaps the contents of two `real(wp)` vectors.
  subroutine osl_swap_vector_wp(a, b)
    real(wp), dimension(:), intent(inout) :: a, b
    real(wp), dimension(size(a)) :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_vector_wp

  !! Swaps the contents of two `integer` vectors.
  subroutine osl_swap_vector_i(a, b)
    integer, dimension(:), intent(inout) :: a, b
    integer, dimension(size(a)) :: temp
    temp = a
    a = b
    b = temp
  end subroutine osl_swap_vector_i

  !--------------------------------------------------------------------------!
  ! Reverse                                                                  !
  !--------------------------------------------------------------------------!

  !! Reverse the elements of a `real(wp)` vector A.
  subroutine osl_reverse_vector_wp(a)
    real(wp), dimension(:), intent(inout) :: a
    real(wp), dimension(size(a, 1)) :: reverse
    integer :: i

    do i = 1, size(a, 1)
       reverse(size(a, 1) - i + 1) = a(i)
    end do
    a = reverse
  end subroutine osl_reverse_vector_wp

  !! Reverse the elements of an `integer` vector A.
  subroutine osl_reverse_vector_i(a)
    integer, dimension(:), intent(inout) :: a
    integer, dimension(size(a, 1)) :: reverse
    integer :: i

    do i = 1, size(a, 1)
       reverse(size(a, 1) - i + 1) = a(i)
    end do
    a = reverse
  end subroutine osl_reverse_vector_i

  !--------------------------------------------------------------------------!
  ! Vector-vector operations                                                 !
  !--------------------------------------------------------------------------!

  !! Return the outer product of the vectors A and B.
  function osl_outer_product(a, b) result(C)
    real(wp), dimension(:), intent(in) :: a
    real(wp), dimension(:), intent(in) :: b
    real(wp), dimension(size(a), size(b)) :: C
    C = spread(a, DIM=2, NCOPIES=size(b)) * spread(b, DIM=1, NCOPIES=size(a))
  end function osl_outer_product

  !--------------------------------------------------------------------------!
  ! Vector operations                                                        !
  !--------------------------------------------------------------------------!

  !! Return N evenly-spaced scalars from A to B.
  function osl_linspace_swp(a, b, n)
    real(wp), intent(in) :: a, b
    integer, intent(in) :: n
    real(wp), dimension(n) :: osl_linspace_swp
    real(wp) :: step
    integer :: i

    step = (b - a) / real(n - 1, wp)
    osl_linspace_swp(1) = a
    do i = 2, n
       osl_linspace_swp(i) = osl_linspace_swp(i-1) + step
    end do
  end function osl_linspace_swp

  !! Return N evenly-spaced vectors from A to B.
  function osl_linspace_vwp(a, b, n)
    real(wp), dimension(:), intent(in) :: a, b
    integer, intent(in) :: n
    real(wp), dimension(n,size(a)) :: osl_linspace_vwp
    real(wp), dimension(size(a)) :: step
    integer :: i

    step = (b - a) / real(n - 1, wp)
    osl_linspace_vwp(1,:) = a
    do i = 2, n
       osl_linspace_vwp(i,:) = osl_linspace_vwp(i-1,:) + step
    end do
  end function osl_linspace_vwp

end module osl_core
