! osl_dist.f90 -- probability distributions
!
! Copyright (C) 2005-2010 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library, a free, open-source
! library for scientific computing, available at http://libosl.org/.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_dist
  use osl_core, wp => osl_wp

  implicit none

  private

  ! Normal Distribution
  public :: osl_normal_pdf
  public :: osl_normal_cdf
  public :: osl_normal_inv

  ! Uniform Distribution
  public :: osl_uniform_pdf
  public :: osl_uniform_cdf
  public :: osl_uniform_inv

  ! Exponential Distribution
  public :: osl_exponential_pdf
  public :: osl_exponential_cdf
  public :: osl_exponential_inv

  ! Gumbel Distribution
  public :: osl_gumbel_pdf
  public :: osl_gumbel_cdf
  public :: osl_gumbel_inv

contains

  !--------------------------------------------------------------------------!
  ! Normal Distribution                                                      !
  !--------------------------------------------------------------------------!

  ! Return the value of the probability distribution function of the
  ! Normal distribution with mean MU and standard deviation SIGMA
  ! evaluated at X.  If either MU or SIGMA are omitted, the respective
  ! default values 0 and 1 will be used.
  function osl_normal_pdf(x, mu, sigma)
    real(wp), intent(in) :: x
    real(wp), intent(in), optional :: mu
    real(wp), intent(in), optional :: sigma
    real(wp) :: osl_normal_pdf
    real(wp) :: z, c

    z = x
    c = 1.0_wp / sqrt(osl_2pi)

    if (present(mu)) then
       z = x - mu
    end if

    if (present(sigma)) then
       z = z / sigma
       c = c / sigma
    end if

    osl_normal_pdf = c * exp( (z**2) / (-2.0_wp) )
  end function osl_normal_pdf

  ! Return the value of the cumulative distribution function of the
  ! Normal distribution with mean MU and standard deviation SIGMA
  ! evaluated at X.  If either MU or SIGMA are omitted, the respective
  ! default values 0 and 1 will be used.
  !
  ! Based on the C++ version of Bernt Odegaard:
  ! http://finance-old.bi.no/~bernt/gcc_prog/
  function osl_normal_cdf(x, mu, sigma)
    real(wp), intent(in) :: x
    real(wp), intent(in), optional :: mu
    real(wp), intent(in), optional :: sigma
    real(wp) :: z
    real(wp) :: osl_normal_cdf
    real(wp) :: a, t, b

    ! A few magical constants
    real(wp), parameter :: b1 =  0.31938153_wp
    real(wp), parameter :: b2 = -0.356563782_wp
    real(wp), parameter :: b3 =  1.781477937_wp
    real(wp), parameter :: b4 = -1.821255978_wp
    real(wp), parameter :: b5 =  1.330274429_wp
    real(wp), parameter :: p  =  0.2316419_wp
    real(wp), parameter :: c2 =  0.3989423_wp

    z = x

    if (present(mu)) then
       z = z - mu
    end if

    if (present(sigma)) then
       z = z / sigma
    end if

    if (z > 6.0_wp) then
       osl_normal_cdf = 1.0_wp
    else if (z < -6.0_wp) then
       osl_normal_cdf = 0.0_wp
    else
       a = abs(z)
       t = 1.0_wp / (1.0_wp + a*p)
       b = c2 * exp((-z)*(z/2.0_wp))
       osl_normal_cdf = ((((b5*t + b4)*t + b3)*t + b2)*t + b1)*t
       osl_normal_cdf = 1.0_wp - b * osl_normal_cdf
       if (z < 0.0_wp) then
          osl_normal_cdf = 1.0_wp - osl_normal_cdf
       end if
    end if
  end function osl_normal_cdf

  ! Return the value of the quantile function, or inverse CDF of the
  ! Normal distribution with mean MU and standard deviation SIGMA
  ! evaluated at P.  If either MU or SIGMA are omitted, the respective
  ! default values 0 and 1 will be used.
  !
  ! Based on the algorithm found here:
  ! http://home.online.no/~pjacklam/notes/invnorm
  function osl_normal_inv(p, mu, sigma)
    real(wp), intent(in) :: p
    real(wp), intent(in), optional :: mu
    real(wp), intent(in), optional :: sigma
    real(wp) :: osl_normal_inv
    real(wp) :: z

    real(wp), parameter :: p_low = 0.02425_wp
    real(wp), parameter :: p_high = 1.0_wp - p_low

    real(wp), parameter, dimension(6) :: a = &
         (/ -3.969683028665376E+01_wp,  2.209460984245205E+02_wp, &
            -2.759285104469687E+02_wp,  1.383577518672690E+02_wp, &
            -3.066479806614716E+01_wp,  2.506628277459239E+00_wp /)
    real(wp), parameter, dimension(5) :: b = &
         (/ -5.447609879822406E+01_wp,  1.615858368580409E+02_wp, &
            -1.556989798598866E+02_wp,  6.680131188771972E+01_wp, &
            -1.328068155288572E+01_wp /)
    real(wp), parameter, dimension(6) :: c = &
         (/ -7.784894002430293E-03_wp, -3.223964580411365E-01_wp, &
            -2.400758277161838E+00_wp, -2.549732539343734E+00_wp, &
             4.374664141464968E+00_wp,  2.938163982698783E+00_wp /)
    real(wp), parameter, dimension(4) :: d = &
         (/  7.784695709041462E-03_wp,  3.224671290700398E-01_wp, &
             2.445134137142996E+00_wp,  3.754408661907416E+00_wp /)

    real(wp) :: q, r

    if (p < p_low) then
       q = sqrt(-2.0_wp * log(p))
       z = (((((c(1)*q+c(2))*q+c(3))*q+c(4))*q+c(5))*q+c(6)) / &
            ((((d(1)*q+d(2))*q+d(3))*q+d(4))*q+1.0_wp)
    else
       if (p <= p_high) then
          q = p - 0.5_wp
          r = q*q
          z = (((((a(1)*r+a(2))*r+a(3))*r+a(4))*r+a(5))*r+a(6))*q / &
               (((((b(1)*r+b(2))*r+b(3))*r+b(4))*r+b(5))*r+1.0_wp)
       else
          if (p < 1) then
             q = sqrt(-2.0_wp * log(1.0_wp - p))
             z = -(((((c(1)*q+c(2))*q+c(3))*q+c(4))*q+c(5))*q+c(6)) / &
                  ((((d(1)*q+d(2))*q+d(3))*q+d(4))*q+1.0_wp)
          else
             z = 0.0_wp
          end if
       end if
    end if

    if (present(sigma)) then
       z = sigma * z
    end if
    if (present(mu)) then
       z = z + mu
    end if

    osl_normal_inv = z
  end function osl_normal_inv

  !--------------------------------------------------------------------------!
  ! Uniform Distribution                                                     !
  !--------------------------------------------------------------------------!

  ! Determine the parameters of the uniform distribution given A and B
  ! (both optional).
  subroutine uniform_parameters(l, u, a, b)
    real(wp), intent(out) :: l
    real(wp), intent(out) :: u
    real(wp), intent(in), optional :: a
    real(wp), intent(in), optional :: b

    if (present(a)) then
       l = a
    else
       l = 0.0_wp
    end if

    if (present(b)) then
       u = b
    else
       u = 1.0_wp
    end if
  end subroutine uniform_parameters

  ! Return the value of the pdf of the Uniform distribution on the
  ! interval [A,B].  If either endpoint A or B is missing, the default
  ! values 0 and 1, respectively, are used.  B must be larger than A.
  real(wp) function osl_uniform_pdf(x, a, b)
    real(wp), intent(in) :: x
    real(wp), intent(in), optional :: a
    real(wp), intent(in), optional :: b
    real(wp) :: l, u

    ! Determine parameters
    call uniform_parameters(l, u, a, b)

    ! Evaluate pdf
    if (x < l .or. x > u) then
       osl_uniform_pdf = 0.0_wp
    else
       osl_uniform_pdf = 1.0_wp / (u - l)
    end if
  end function osl_uniform_pdf

  ! Return the value of the cdf of the Uniform distribution on the
  ! interval [A,B].  If either endpoint A or B is missing, the default
  ! values 0 and 1, respectively, are used.  B must be larger than A.
  real(wp) function osl_uniform_cdf(x, a, b)
    real(wp), intent(in) :: x
    real(wp), intent(in), optional :: a
    real(wp), intent(in), optional :: b
    real(wp) :: l, u

    ! Determine parameters
    call uniform_parameters(l, u, a, b)

    ! Evaluate cdf
    if (x < l) then
       osl_uniform_cdf = 0.0_wp
    else if (x > u) then
       osl_uniform_cdf = 1.0_wp
    else
       osl_uniform_cdf = x / (u - l)
    end if
  end function osl_uniform_cdf

  ! Return the value of the inverse cdf, or quantile function, of the
  ! Uniform distribution on the interval [A,B].  If either endpoint A
  ! or B is missing, the default values 0 and 1, respectively, are
  ! used.  B must be larger than A.  P must be such that 0 <= P <= 1.
  ! Otherwise, the function returns zero.
  real(wp) function osl_uniform_inv(p, a, b)
    real(wp), intent(in) :: p
    real(wp), intent(in), optional :: a
    real(wp), intent(in), optional :: b
    real(wp) :: l, u

    ! Determine parameters
    call uniform_parameters(l, u, a, b)

    ! Evaluate inverse cdf
    if (p < 0.0_wp .or. p > 1.0_wp) then
       osl_uniform_inv = 0.0_wp
    else
       osl_uniform_inv = p * (u - l)
    end if
  end function osl_uniform_inv

  !--------------------------------------------------------------------------!
  ! Exponential Distribution                                                 !
  !--------------------------------------------------------------------------!

  real(wp) function osl_exponential_pdf(x, lambda)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: lambda

    if (x >= 0.0_wp) then
       osl_exponential_pdf = lambda * exp(-lambda * x)
    else
       osl_exponential_pdf = 0.0_wp
    end if
  end function osl_exponential_pdf

  real(wp) function osl_exponential_cdf(x, lambda)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: lambda

    if (x >= 0.0_wp) then
       osl_exponential_cdf = 1.0_wp - exp(-lambda * x)
    else
       osl_exponential_cdf = 0.0_wp
    end if
  end function osl_exponential_cdf

  real(wp) function osl_exponential_inv(p, lambda)
    real(wp), intent(in) :: p
    real(wp), intent(in) :: lambda

    if (p >= 0.0_wp .and. p < 1.0_wp) then
       osl_exponential_inv = -1.0_wp * log(1.0_wp - p) / lambda
    else
       osl_exponential_inv = 0.0_wp
    end if
  end function osl_exponential_inv

  !--------------------------------------------------------------------------!
  ! Gumbel Distribution (Type I Extreme Value Distribution)                  !
  !--------------------------------------------------------------------------!

  !! Return the value of the PDF of the Gumbel distribution with
  !! parameters MU and BETA at X.  The default values for MU and BETA
  !! are `0.0_wp` and `1.0_wp`, respectively, corresponding to the
  !! standard Gumbel distribution.
  function osl_gumbel_pdf(x, mu, beta)
    real(wp), intent(in) :: x
    real(wp), intent(in), optional :: mu
    real(wp), intent(in), optional :: beta
    real(wp) :: osl_gumbel_pdf
    real(wp) :: m, b, z

    if (present(mu)) then
       m = mu
    else
       m = 0.0_wp
    end if

    if (present(beta)) then
       b = beta
    else
       b = 1.0_wp
    end if

    z = exp(-(x - m) / b)
    osl_gumbel_pdf = z * exp(-z) / b
  end function osl_gumbel_pdf

  !! Return the value of the CDF of the Gumbel distribution with
  !! parameters MU and BETA at X.  The default values for MU and BETA
  !! are `0.0_wp` and `1.0_wp`, respectively, corresponding to the
  !! standard Gumbel distribution.
  function osl_gumbel_cdf(x, mu, beta)
    real(wp), intent(in) :: x
    real(wp), intent(in), optional :: mu
    real(wp), intent(in), optional :: beta
    real(wp) :: osl_gumbel_cdf
    real(wp) :: m, b

    if (present(mu)) then
       m = mu
    else
       m = 0.0_wp
    end if

    if (present(beta)) then
       b = beta
    else
       b = 1.0_wp
    end if

    osl_gumbel_cdf = exp(-exp(-(x - m)/b))
  end function osl_gumbel_cdf

  !! Return the value of the inverse CDF of the Gumbel distribution
  !! with parameters MU and BETA at X.  The default values for MU and
  !! BETA are `0.0_wp` and `1.0_wp`, respectively, corresponding to
  !! the standard Gumbel distribution.
  function osl_gumbel_inv(p, mu, beta)
    real(wp), intent(in) :: p
    real(wp), intent(in), optional :: mu
    real(wp), intent(in), optional :: beta
    real(wp) :: osl_gumbel_inv
    real(wp) :: m, b

    if (present(mu)) then
       m = mu
    else
       m = 0.0_wp
    end if

    if (present(beta)) then
       b = beta
    else
       b = 1.0_wp
    end if

    if (p > 0.0_wp .and. p < 1.0_wp) then
       osl_gumbel_inv = m - b * log(-log(p))
    else
       osl_gumbel_inv = 0.0_wp
    end if
  end function osl_gumbel_inv

end module osl_dist
